package ap.proof;

import ap.terfor.{Formula, TermOrder}
import ap.terfor.conjunctions.{Conjunction, Quantifier}
import ap.proof.goal._
import ap.util.{Logic, Debug}
import ap.parameters.{GoalSettings, Param}

object ExhaustiveProver {
  
  private def AC = Debug.AC_PROVER
  
}

/**
 * A prover that tries to construct an exhaustive proof for a given goal. The
 * prover tries to optimise by early stopping the expansion of the proof tree
 * if it is detected that a certain subtree can never yield a satisfiable
 * closing constraint
 */
class ExhaustiveProver(simplifier : ConstraintSimplifier,
                       stoppingCond : (ProofTree) => Boolean,
                       depthFirst : Boolean,
                       settings : GoalSettings) {

  def this(simplifier : ConstraintSimplifier) =
    this(simplifier, (tree) => false, false, GoalSettings.DEFAULT)
   
  private def ptfStoppingCond(goal : Goal) =
    stoppingCond(goal) ||
    (goal.tasks.max match {
 //   case FactsNormalisationTask | EliminateFactsTask |
 //        _ : AddFactsTask | _ : NegLitClauseTask => false
      case FactsNormalisationTask | EliminateFactsTask | UpdateTasksTask => false
      case task : AddFactsTask => Param.POS_UNIT_RESOLUTION(goal.settings) &&
                                  !task.formula.predicates.isEmpty
      case task : BetaFormulaTask => task.splittingNecessary(goal)
      case task : DivisibilityTask => task.splittingNecessary(goal)
      case SplitEqTask => SplitEqTask.splittingNecessary(goal)
      case OmegaTask => OmegaTask.splittingNecessary(goal)
      case _ => true
    })

  private val ptf =
    new IteratingProofTreeFactory(ptfStoppingCond _, simplifier)
   
  //////////////////////////////////////////////////////////////////////////////

  def apply(inputFor : Formula, order : TermOrder) : ProofTree = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(ExhaustiveProver.AC,
                    inputFor.variables.isEmpty && (order isSortingOf inputFor))
    ////////////////////////////////////////////////////////////////////////////
    val goal = Goal(inputFor, order, settings)
    (if (depthFirst)
       expandDepthFirstUntilSat(goal, false)
     else
       expandFairUntilSat(goal, false)) _1
  }
       
  private def continueProving(tree : ProofTree) =
    tree.stepMeaningful && !stoppingCond(tree) &&
    // when proving in depth-first mode, a tree is expanded only as long as
    // its constraint is not satisfiable
    (!depthFirst || ModelSearchProver(tree.closingConstraint.negate,
                                      tree.closingConstraint.order).isFalse)
  
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Expand the proof breadth-first until the root constraint is satisfiable
   * (or until <code>stoppingCond</code> returns <code>true</code>).
   * The result is a pair consisting of the new proof tree and a boolean
   * that tells whether it was possible to apply any steps
   */
  private def expandFairUntilSat(_tree : ProofTree, swpBefore : Boolean)
                                                     : (ProofTree, Boolean) = {
    var tree : ProofTree = _tree
    var cont : Boolean = true
    var swp : Boolean = swpBefore
    while (cont && continueProving(tree)) {
  /*  println(tree)
     println(goalNum(tree))
     println  */
      val (newTree, newCont) = expandProofGoals(tree, 1)
      tree = newTree
      cont = newCont
      if (newCont) swp = true
    }
    (tree, swp)
  }
  
  private def isGoalLike(tree : ProofTree) : Boolean = tree match {
    case _ : Goal                                => true
    case QuantifiedTree(Quantifier.ALL, subtree) => isGoalLike(subtree)
    case SubstTree(_, subtree)                   => isGoalLike(subtree)
    case _                                       => false
  }
  
  /**
   * Expand the proof depth-first until the root constraint is satisfiable
   * (or until <code>stoppingCond</code> returns <code>true</code>).
   * The result is a pair consisting of the new proof tree and a boolean
   * that tells whether it was possible to apply any steps
   */
  private def expandDepthFirstUntilSat(tree : ProofTree, swpBefore : Boolean)
                                                       : (ProofTree, Boolean) =
    if (continueProving(tree))
      tree match {
      case _ : ProofTree if (isGoalLike(tree)) => {
        // we always consider the goal together with the enclosing universal
        // quantifiers and substitutions
        val (newTree, swp) = expandProofGoals(tree, 1)
        
        if (swp) {
          // recurse because the structure of the tree might have changed
          expandDepthFirstUntilSat(newTree, true)
        } else {
          // we continue in a fair manner
          expandFairUntilSat(newTree, swpBefore)
        }        
      }
      
      case goal : Goal => expandDepthFirstUntilSat(goal step ptf, true)
      
      case tree : WeakenTree => // should not happen
                                { assert(false); null }
      
      case tree : ProofTreeOneChild => {
        val (newSubtree, swp) = expandDepthFirstUntilSat(tree.subtree, false)
        val newTree = tree update newSubtree
        
        if (swp) {
          // recurse because the structure of the tree might have changed
          expandDepthFirstUntilSat(newTree, true)
        } else {
          // we continue in a fair manner
          expandFairUntilSat(newTree, swpBefore)
        }        
      }
      
      case tree : AndTree => {
        val (newLeft, leftSWP) = expandDepthFirstUntilSat(tree.left, false)
        val (newRight, rightSWP) = expandDepthFirstUntilSat(tree.right, false)
        
        val newTree = tree.update(newLeft, newRight)
        
        if (leftSWP || rightSWP) {
          // recurse because the structure of the tree might have changed
          expandDepthFirstUntilSat(newTree, true)
        } else {
          // we continue in a fair manner
          expandFairUntilSat(newTree, swpBefore)
        }
      }
    } else {
      // otherwise, the tree is already satisfiable (or at least the constraint
      // is not false .. TODO) or no more steps are possible
      (tree, swpBefore)
    }
  
  
  /**
   * Try to expand each goal of <code>tree</code> by applying <code>limit</code>
   * steps. The result is a pair consisting of the new proof tree and a boolean
   * that tells whether it was possible to apply any steps
   */
  private def expandProofGoals(tree : ProofTree, limit : Int)
                                                     : (ProofTree, Boolean) =
    if (limit > 0 && tree.stepMeaningful) tree match {
      
      case goal : Goal => (expandProofGoals(goal.step(ptf), limit - 1) _1, true)
      
      case tree : ProofTreeOneChild => {
        val (newSubtree, stepWasPossible) = expandProofGoals(tree.subtree, limit)
        (tree update newSubtree, stepWasPossible)
      }
      
      case tree : AndTree => {
        val (newLeft, leftSWP) = expandProofGoals(tree.left, limit)
        val (newRight, rightSWP) = expandProofGoals(tree.right, limit)        
        (tree.update(newLeft, newRight), leftSWP || rightSWP)
      }
      
    } else {
      (tree, false)
    }
   
   private def goalNum(tree : ProofTree) : Int = tree match {
     case g : Goal => if (g.closingConstraint.isTrue) 0 else 1
     case t : ProofTreeOneChild => goalNum(t.subtree)
     case t : AndTree => goalNum(t.left) + goalNum(t.right)
   }
}
