package ap.proof;

import ap.basetypes.IdealInt
import ap.terfor.{Formula, TermOrder, ConstantTerm}
import ap.terfor.conjunctions.{Conjunction, Quantifier}
import ap.terfor.linearcombination.LinearCombination
import ap.terfor.equations.EquationConj
import ap.terfor.substitutions.{Substitution, IdentitySubst}
import ap.proof.goal.{Goal, FactsNormalisationTask, EliminateFactsTask,
                      AddFactsTask, UpdateTasksTask}
import ap.util.{Debug, Logic, LRUCache}
import ap.parameters.GoalSettings

import scala.collection.mutable.Stack


/**
 * A prover that tries to construct a countermodel of a ground formula. This
 * prover works in depth-first mode, in contrast to the
 * <code>ExhaustiveProver</code>.
 */
object ModelSearchProver {

  private val AC = Debug.AC_PROVER
   
  private val simplifier = ConstraintSimplifier.NO_SIMPLIFIER
  
  // we need to store eliminated facts from goals, otherwise we could not
  // construct a complete model
  private val ptf = new SimpleProofTreeFactory(true, simplifier) {
    override def eliminatedConstant(subtree : ProofTree,
                                    c : ConstantTerm,
                                    witness : (Substitution, TermOrder) => Substitution,
                                    order : TermOrder) : ProofTree =
      new WitnessTree (subtree, c, witness, order)
  }

  private val cache = new LRUCache[Formula, Conjunction] (1000)
  
  abstract class Result
  case class FoundModel(model : Conjunction) extends Result
  case object NoModel extends Result
  case object SearchCancelled extends Result
  
  /**
   * <code>inputFor</code> is the formula to be disproven. The result of the
   * method is a countermodel of <code>inputFor</code>, or <code>FALSE</code>
   * if it was not possible to find one (this implies that <code>inputFor</code>
   * is valid).
   */
  def apply(inputFor : Formula, order : TermOrder) : Conjunction = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(ModelSearchProver.AC,
                    inputFor.variables.isEmpty && (order isSortingOf inputFor))
    ////////////////////////////////////////////////////////////////////////////
    cache(inputFor) {
      applyHelp(inputFor, order, false, GoalSettings.DEFAULT) match {
        case FoundModel(m) => m
        case NoModel => Conjunction.FALSE
      }
    }
  }
   
  def apply(inputFor : Formula, order : TermOrder,
            stoppingCond : => Boolean,
            settings : GoalSettings) : Result = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(ModelSearchProver.AC,
                    inputFor.variables.isEmpty && (order isSortingOf inputFor))
    ////////////////////////////////////////////////////////////////////////////
    applyHelp(inputFor, order, stoppingCond, settings)
  }
   
  /**
   * <code>inputFor</code> is the formula to be disproven. The result of the
   * method is a countermodel of <code>inputFor</code>, or <code>FALSE</code>
   * if it was not possible to find one (this implies that <code>inputFor</code>
   * is valid).
   */
  private def applyHelp(inputFor : Formula, order : TermOrder,
                        stoppingCond : => Boolean,
                        settings : GoalSettings) : Result = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(ModelSearchProver.AC, inputFor.variables.isEmpty)
    ////////////////////////////////////////////////////////////////////////////
    val goal = Goal(inputFor, Set() ++ order.orderedConstants, order, settings)
    try {
      expandProof(goal, stoppingCond, 0)
      NoModel
    } catch {
      case StoppedSearchExc => SearchCancelled
        
      case FoundModelExc(facts, constantValues, order) => {
        val valuesConj =
          EquationConj(for (c <- order.orderedConstants.elements)
                       yield LinearCombination(Array((IdealInt.ONE, c),
                                                     (IdealInt.MINUS_ONE, constantValues(c))),
                                               order),
                       order)
        FoundModel(Conjunction.conj(Array(valuesConj, facts), order))
      }
    }
  }

  private def isSatisfiableGoal(tree : ProofTree) : Boolean = tree match {
    case goal : Goal => {
      val c = goal.closingConstraint
      if (c.isTrue) {
        true
      } else if (c.isFalse) {
        false
      } else {
        // this case can only occur if the proof contains uninterpreted
        // predicates; we have to check the constraint recursively
        // we only do that if the next rule application would be non-trivial
        (goal.tasks.isEmpty ||
         (goal.tasks.max match {
          case FactsNormalisationTask | EliminateFactsTask | UpdateTasksTask => false
          case task : AddFactsTask => !task.formula.predicates.isEmpty
          case _ => true
          })) &&
        ModelSearchProver(c, goal.order).isFalse
      }
    }
    case _ => false
  }
   
  private def expandProof(tree : ProofTree, stoppingCond : => Boolean,
                          depth : Int) : unit = {
    if (stoppingCond) throw StoppedSearchExc
      
    if (isSatisfiableGoal(tree)) {
 //     println("backtracking (depth " + depth + ")")
    } else
    tree match { 
      case goal : Goal => {
        if (goal.stepPossible)
          expandProof(goal.step(ptf), stoppingCond, depth)
        else
          throw FoundModelExc(goal.facts,
                              new IdentitySubst (goal.order),
                              goal.order)
      }

      case tree : WitnessTree => {
        // we build a local stack with all witnesses to avoid
        // StackOverflowExceptions
        val witnessStack = new Stack[(Substitution, TermOrder) => Substitution]
                                      
        var subtree : ProofTree = tree
        while (subtree.isInstanceOf[WitnessTree]) {
          val tree = subtree.asInstanceOf[WitnessTree]
          witnessStack push tree.witness
          subtree = tree.subtree
        }

        try {
          expandProof(subtree, stoppingCond, depth)
          ////////////////////////////////////////
          // some arbitrary operations to avoid tail recursion ...
          // this seems to be a compiler bug
          // TODO: write a bug report
          var x = 0; x = x
        } catch {
          case FoundModelExc(facts, constantValues, order) => {
            var newCValues = constantValues
            while (!witnessStack.isEmpty)
              newCValues = witnessStack.pop()(newCValues, order)
            
            throw FoundModelExc(facts, newCValues, order)
          }
        }
      }
      
      case tree : ProofTreeOneChild => {
        ////////////////////////////////////////////////////////////////////////
        Debug.assertInt(ModelSearchProver.AC, tree match {
                          case _ : WeakenTree => false
                          case tree : QuantifiedTree => tree.quan == Quantifier.ALL
                          case _ => true
                        })
        ////////////////////////////////////////////////////////////////////////
        expandProof(tree.subtree, stoppingCond, depth)
      }
     
      case tree : AndTree => {
        expandProof(tree.left, stoppingCond, depth + 1)
        expandProof(tree.right, stoppingCond, depth + 1)
      }
      
    }
  }
  
  private case class FoundModelExc(facts : Conjunction,
                                   constantValues : Substitution,
                                   order : TermOrder) extends Exception

  private case object StoppedSearchExc extends Exception
}

 
private case class WitnessTree(val subtree : ProofTree,
                               val eliminatedConstant : ConstantTerm,
                               val witness : (Substitution, TermOrder) => Substitution,
                               val order : TermOrder) extends ProofTreeOneChild {
  
  def update(newSubtree : ProofTree) : ProofTree =
    new WitnessTree(subtree, eliminatedConstant, witness, order)

  lazy val rawClosingConstraint : Conjunction = subtree.closingConstraint

  lazy val closingConstraint : Conjunction = subtree.closingConstraint

  lazy val stepMeaningful : Boolean = subtree.stepMeaningful

  protected val subtreeOrder = order
  
}