package ap.proof.goal;

import ap.terfor.{Term, Formula, TermOrder, ConstantTerm, VariableTerm}
import ap.terfor.conjunctions.{Conjunction, Quantifier, ReduceWithConjunction,
                               NegatedConjunctions}
import ap.terfor.linearcombination.LinearCombination
import ap.basetypes.IdealInt
import ap.terfor.arithconj.ArithConj
import ap.terfor.equations.{ColumnSolver, NegEquationConj, EquationConj}
import ap.terfor.substitutions.{ConstantSubst, PseudoConstantSubst}
import ap.util.{Logic, Debug, Seqs}

case object FactsNormalisationTask extends EagerTask {

  val AC = Debug.AC_FACTS_TASK
  
  def apply(goal : Goal, ptf : ProofTreeFactory) : ProofTree = {
    
    var facts = goal.facts
    var eliminatedConstants = goal.eliminatedConstants
    var postProcessor : ProofTree => ProofTree = ((p) => p)
    var order = goal.order
    var iteration = 0
    
    def positiveEqsSolvable : Boolean =
      Logic.exists(for (lc <- facts.arithConj.positiveEqs.elements)
                   yield !lc.leadingCoeff.isOne)
    
    def solvePositiveEqs : unit = {
      val solver = new GoalColumnSolver (facts.arithConj.positiveEqs, eliminatedConstants,
                                         "sc_" + goal.age + "_" + iteration,
                                         order, ptf)
      val (newEqConj, newOrder) = solver.result
  
      order = newOrder
      facts = facts.updatePositiveEqs(newEqConj, order)
      eliminatedConstants = solver.eliminatedConstants
      postProcessor = postProcessor compose solver.postProcessor      
    }

    if (positiveEqsSolvable) solvePositiveEqs

    var cont : Boolean = true
    while (cont) {
      // propagate the solved equations into the other facts
      facts = ReduceWithConjunction(Conjunction.TRUE, order)(facts)

      if (facts.isFalse)
        // then the goal can be closed immediately
        return ptf.updateGoal(Conjunction.FALSE, goal)

      iteration = iteration + 1
      // if the equations have changed, we might have to solve them again
      if (positiveEqsSolvable)
        solvePositiveEqs
      else
        cont = false
    }

    // update clauses
    
    var newClauses = ReduceWithConjunction(facts, order)(goal.clauses)
      
    val (otherStuff, realClauses) =
      Seqs.split(newClauses.elements,
                 (c:Conjunction) => (c.isLiteral || c.isNegatedConjunction))
    newClauses = newClauses.update(realClauses, order)

    val newTasks = for (c <- otherStuff; t <- goal.formulaTasks(c)) yield t

    if (newTasks.isEmpty && facts == goal.facts && newClauses == goal.clauses)
      // nothing has changed and the application of this task was unnecessary
      ptf.updateGoal(goal)
    else
      postProcessor(ptf.updateGoal(facts, newClauses, eliminatedConstants, order,
                                   newTasks, goal))
  }

}


////////////////////////////////////////////////////////////////////////////////

private class GoalColumnSolver(eqs : EquationConj,
                               var eliminatedConstants : Set[ConstantTerm],
                               constBasename : String,
                               order : TermOrder, ptf : ProofTreeFactory)
              extends ColumnSolver(eqs, order) {
  
  private var constCounter : Int = 0
  
  /**
   * The function that is supposed to be applied to the resulting proof tree
   * (adding the <code>QuantifiedTree</code>, <code>SubstTree</code>
   * constructors)
   */
  var postProcessor : ProofTree => ProofTree = ((p) => p)
  
  //////////////////////////////////////////////////////////////////////////////
  
  protected def isSolvableEq(lc : LinearCombination, order : TermOrder) =
    if (lc.leadingCoeff.isOne) {
      None
    } else if (noneOrTwoElimConstants(lc)) {
      // normal column operation
      Some(reduceWithLeadingCoeff(lc, order))
    } else {
      // then lc contains exactly one eliminated constant, which is the leading
      // term and has a coefficient that is not 1
      Debug.assertInt(FactsNormalisationTask.AC,
                      {
                        val (c, t) = lc(0)
                        !c.isOne && isEliminated(t)
                      } &&
                      Logic.forall(1, lc.size, (i) => !isEliminated(lc(i) _2)))
      Some(makeLeadingTermSmall(lc, order))
    }

  private def newConst : ConstantTerm = {
    val res = new ConstantTerm(constBasename + "_" + constCounter)
    constCounter = constCounter + 1
    res
  }
  
  /**
   * Introduce a small constant as new name for the leading term of
   * <code>lc</code> (rule div-close)
   */
  private def makeLeadingTermSmall(lc : LinearCombination,
                                   order : TermOrder)
                                      : (Term, LinearCombination, TermOrder) = {
    val (leadingCoeff, leadingTerm) = lc(0)
    val smallConst = newConst
    val extendedOrder = order.extend(smallConst, order.orderedConstants)
    
    // the new constant (times the old leading coefficient) have to be
    // substituted is the closing constraint
    val symDefinition = LinearCombination(leadingTerm, order)
    val substLC = LinearCombination.sum(Array((leadingCoeff, symDefinition),
                                              (IdealInt.MINUS_ONE, lc)),
                                        order)    
    val backSubst = new PseudoConstantSubst(leadingCoeff, smallConst,
                                            substLC, extendedOrder)
    
    // the negated divisibility judgement has to be added disjunctively
    // to the closing constraint
    val var0 = LinearCombination(VariableTerm(0), order)
    val lcMod = LinearCombination.sum(Array((leadingCoeff, var0),
                                            (-leadingCoeff, symDefinition),
                                            (IdealInt.ONE, lc)),
                                      order)
    val notDividedByLC = Conjunction.quantify(Array(Quantifier.ALL),
                                              NegEquationConj(lcMod, order),
                                              order)
                                            
    postProcessor = postProcessor compose
                    ((p:ProofTree) =>
                          ptf.weaken(ptf.substitute(p, backSubst, order, extendedOrder),
                                     notDividedByLC,
                                     order))
    (smallConst, symDefinition, extendedOrder)
  }
  
  /**
   * Introduce a new small constant and reduce the coefficients of
   * <code>lc</code> with the leading coefficient (rules col-red, col-red-subst)
   */
  private def reduceWithLeadingCoeff(lc : LinearCombination,
                                     order : TermOrder)
                                      : (Term, LinearCombination, TermOrder) = {
    val symDefinition = lc.reduceWithLeadingCoeff
    val smallConst = newConst
    
    if (isEliminated(lc(0) _2)) {
      // then also the new constant can be eliminated, and has to be put in
      // between the non-eliminated and the eliminated constants
      val extendedOrder = order.extend(smallConst, eliminatedConstants)
      eliminatedConstants = eliminatedConstants + smallConst
      postProcessor = postProcessor compose
                      ((p:ProofTree) =>
                       ptf.quantify(p, Quantifier.ALL, Set(smallConst),
                                    order, extendedOrder))
      (smallConst, symDefinition, extendedOrder)
    } else {
      // then it is not possible to eliminate the new constant, and the constant
      // can be made smaller than all other symbols
      val extendedOrder = order.extend(smallConst, order.orderedConstants)
      val backSubst = new ConstantSubst(smallConst, symDefinition, extendedOrder)
      postProcessor = postProcessor compose
                      ((p:ProofTree) => ptf.substitute(p, backSubst, order, extendedOrder))
      (smallConst, symDefinition, extendedOrder)
    }
  }

  //////////////////////////////////////////////////////////////////////////////
  
  private def isEliminated(t : Term) : Boolean = t match {
    case t : ConstantTerm => eliminatedConstants contains t
    case _ => false
  }

  /**
   * Determine whether the linear combination <code>lc</code> contains either
   * none or at least two terms that are eliminated in <code>goal</code>
   */
  private def noneOrTwoElimConstants(lc : LinearCombination) : Boolean = {
    ////////////////////////////////////////////////////////////////////////////
    def post(b : Boolean) = {
      Debug.assertPost(FactsNormalisationTask.AC,
                       b ==
                       (Logic.exists(0, lc.size, (i) =>
                        Logic.exists(i+1, lc.size, (j) =>
                          isEliminated(lc(i) _2) && isEliminated(lc(j) _2))))
                        ||
                        Logic.forall(for ((_, t) <- lc.elements)
                                     yield !isEliminated(t)))
      b
    }
    ////////////////////////////////////////////////////////////////////////////

    post (if (lc.isEmpty) {
            true
          } else if (isEliminated(lc(0) _2)) {
            (lc.size >= 2) && isEliminated(lc(1) _2)
          } else {
            true
          })
  }
}
