package ap.terfor.arithconj;

import ap.terfor.linearcombination.LinearCombination
import ap.terfor.conjunctions.{Conjunction, NegatedConjunctions}
import ap.terfor.equations.{EquationConj, NegEquationConj, ReduceWithEqs,
                            ReduceWithNegEqs}
import ap.terfor.inequalities.{InEqConj, ReduceWithInEqs}
import ap.terfor.preds.{Atom, PredConj}
import ap.util.{Debug, Seqs, FilterIt}

object ReduceWithAC {
  
  private val AC = Debug.AC_PROPAGATION

  def apply(ac : ArithConj, order : TermOrder) : ReduceWithAC = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(AC, ac isSortedBy order)
    ////////////////////////////////////////////////////////////////////////////
    
    new ReduceWithAC(ReduceWithEqs(ac.positiveEqs, order),
                     ReduceWithNegEqs(ac.negativeEqs, order),
                     ReduceWithInEqs(ac.inEqs, order),
                     order)
  }

  //////////////////////////////////////////////////////////////////////////////
  // Some of the "static" methods of the <code>ReduceWithAC</code>-class
  // These are methods juggling with different reducer-objects
      
  /**
   * Reduce a conjunction of arithmetic stuff and return the reduced conjunction,
   * together with a new <code>ReduceWithAC</code> object to which the reduced
   * conjunction was added.
   */
  private def reduceAC(ac : ArithConj,
                       initialReducer : ReduceWithAC) : (ArithConj, ReduceWithAC) = {

    // positive equations always come first
    val newPosEqs = initialReducer.reduce(ac.positiveEqs, ac.inEqs.equalityInfs)
    var reducer = initialReducer addEquations newPosEqs
      
    // then, reduce inequalities, assuming the (unreduced) negated equations
    val newInEqs = (reducer addEquations ac.negativeEqs) reduce ac.inEqs
    reducer = reducer addInEqs newInEqs
      
    // reduce negated equations, assuming the reduced inequalities
    val newNegEqs = reducer reduce ac.negativeEqs
    reducer = reducer addEquations newNegEqs
      
    val newAC = ArithConj(newPosEqs, newNegEqs, newInEqs, reducer.order)
    if (newInEqs.equalityInfs.isEmpty && newNegEqs == ac.negativeEqs)
      (newAC, reducer)
    else
      // if the new inequalities still imply equations, we have to reduce once
      // more. note, that we again start with the reducer <code>initialReducer</code> 
      reduceAC(newAC, initialReducer)
  }

}

class ReduceWithAC private (positiveEqs : ReduceWithEqs,
                            negativeEqs : ReduceWithNegEqs,
                            inEqs : ReduceWithInEqs,
                            private val order : TermOrder) {

  def passQuantifiers(num : Int) : ReduceWithAC = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(ReduceWithAC.AC, num >= 0)
    ////////////////////////////////////////////////////////////////////////////
    if (num == 0)
      this
    else
      new ReduceWithAC(positiveEqs passQuantifiers num,
                       negativeEqs passQuantifiers num,
                       inEqs passQuantifiers num,
                       order)
  }

  def addArithConj(ac : ArithConj) : ReduceWithAC = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(ReduceWithAC.AC, ac isSortedBy order)
    ////////////////////////////////////////////////////////////////////////////
    if (ac.isEmpty)
      this
    else
      new ReduceWithAC(positiveEqs addEquations ac.positiveEqs.toMap,
                       negativeEqs addEquations ac.negativeEqs.toSet,
                       inEqs addInEqs ac.inEqs,
                       order)    
  }
  
  private def addEquations(eqs : EquationConj) : ReduceWithAC = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(ReduceWithAC.AC, eqs isSortedBy order)
    ////////////////////////////////////////////////////////////////////////////
    if (eqs.isEmpty)
      this
    else
      new ReduceWithAC(positiveEqs addEquations eqs.toMap,
                       negativeEqs,
                       inEqs,
                       order)
  }
  
  private def addEquations(eqs : NegEquationConj) : ReduceWithAC = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(ReduceWithAC.AC, eqs isSortedBy order)
    ////////////////////////////////////////////////////////////////////////////
    if (eqs.isEmpty)
      this
    else
      new ReduceWithAC(positiveEqs,
                       negativeEqs addEquations eqs.toSet,
                       inEqs,
                       order)
  }
  
  private def addInEqs(furtherInEqs : InEqConj) : ReduceWithAC = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(ReduceWithAC.AC, furtherInEqs isSortedBy order)
    ////////////////////////////////////////////////////////////////////////////
    if (furtherInEqs.isEmpty)
      this
    else
      new ReduceWithAC(positiveEqs,
                       negativeEqs,
                       inEqs addInEqs furtherInEqs,
                       order)    
  }
  
  //////////////////////////////////////////////////////////////////////////////
  // Methods for reducing different kinds of formulas. If any of the
  // methods detects that a resulting formula is false,
  // <code>FALSE_EXCEPTION</code> is thrown (this is done to simplify the
  // handling of <code>Conjunction</code>s)
  // TODO: optimise for cases where no reduction is possible (detect this early,
  // create no new objects)
  // TODO: define the following methods in a nicer way, polymorphic?
  
  private def reduce(eqs1 : EquationConj, eqs2 : EquationConj) : EquationConj =
    if (eqs2.isTrue)
      reduce(eqs1)
    else if (eqs1.isTrue)
      reduce(eqs2)
    else
      reduce(EquationConj.conj(Array(eqs1, eqs2), order))

  private def reduce(eqs : EquationConj) : EquationConj = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(ReduceWithAC.AC, eqs isSortedBy order)
    ////////////////////////////////////////////////////////////////////////////
    val redEqs = inEqs(negativeEqs(positiveEqs(eqs)))
    if (redEqs.isFalse) throw FALSE_EXCEPTION_STD
    redEqs
  }

  private def reduce(eqs : NegEquationConj) : NegEquationConj = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(ReduceWithAC.AC, eqs isSortedBy order)
    ////////////////////////////////////////////////////////////////////////////
    val redEqs = inEqs(negativeEqs(positiveEqs(eqs)))
    if (redEqs.isFalse) throw FALSE_EXCEPTION_STD
    redEqs
  }

  private def reduce(ies : InEqConj) : InEqConj = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(ReduceWithAC.AC, ies isSortedBy order)
    ////////////////////////////////////////////////////////////////////////////
    val redInEqs = inEqs(negativeEqs(positiveEqs(ies)))
    if (redInEqs.isFalse) throw FALSE_EXCEPTION_STD
    redInEqs
  }

  private def reduce(conj : PredConj) : PredConj = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(ReduceWithAC.AC, conj isSortedBy order)
    ////////////////////////////////////////////////////////////////////////////
    val redConj = positiveEqs(conj)
    if (redConj.isFalse) throw FALSE_EXCEPTION_PRED(redConj)
    redConj
  }

  //////////////////////////////////////////////////////////////////////////////
    
  /**
   * Reduce an arithmetic conjunction using the information stored in this
   * object. The result is the simplified conjunction, as well as a new
   * reducer to which the information from the simplified arithmetic conjunction
   * has been added.
   */
  def reduceAndAdd(conj : ArithConj) : (ArithConj, ReduceWithAC) =
    if (conj.isTrue || conj.isFalse)
      (conj, this)
    else
      try { ReduceWithAC.reduceAC(conj, this) }
      catch { case _ : FALSE_EXCEPTION => (ArithConj.FALSE, this) }

  def apply(conj : ArithConj) : ArithConj =  {
    val res = (reduceAndAdd(conj) _1)
    ////////////////////////////////////////////////////////////////////////////
    // we demand that the reducer is a projection (repeated application does not
    // change the result anymore)
    Debug.assertPost(ReduceWithAC.AC, (reduceAndAdd(res) _1) == res)
    ////////////////////////////////////////////////////////////////////////////
    res
  }
     
  def apply(conj : EquationConj) : EquationConj =
    try { this reduce conj }
    catch { case _ : FALSE_EXCEPTION => EquationConj.FALSE }    
    
  def apply(conj : PredConj) : PredConj =
    if (positiveEqs.isEmpty)
      conj
    else
      try { this reduce conj }
      catch { case FALSE_EXCEPTION_PRED(f) => f }    
    
}

private abstract class FALSE_EXCEPTION extends Exception

private object FALSE_EXCEPTION_STD extends FALSE_EXCEPTION 

private case class FALSE_EXCEPTION_PRED(conj : PredConj) extends FALSE_EXCEPTION
